import os.path

from .EmbPattern import EmbPattern
import pyembroidery.DstWriter as DstWriter
import pyembroidery.PecWriter as PecWriter
import pyembroidery.PesWriter as PesWriter
import pyembroidery.ExpWriter as ExpWriter
import pyembroidery.Vp3Writer as Vp3Writer
import pyembroidery.JefWriter as JefWriter
import pyembroidery.SvgWriter as SvgWriter
import pyembroidery.DstReader as DstReader
import pyembroidery.PecReader as PecReader
import pyembroidery.PesReader as PesReader
import pyembroidery.ExpReader as ExpReader
import pyembroidery.Vp3Reader as Vp3Reader
import pyembroidery.JefReader as JefReader


def convert(filename_from, filename_to, encode_settings=None):
    pattern = read(filename_from)
    if pattern is None:
        return
    pattern = pattern.get_stable_pattern()
    write(pattern, filename_to, encode_settings)


def get_extension_by_filename(filename):
    """extracts he extension from a filename"""
    return os.path.splitext(filename)[1][1:]


def read_embroidery(reader, f, pattern=None):
    """Reads fileobject or filename with reader."""
    if pattern is None:
        pattern = EmbPattern()
    if isinstance(f, str):
        with open(f, "wb") as stream:
            reader.read(stream, pattern)
    else:
        reader.read(f, pattern)
    return pattern


def read_dst(f, pattern=None):
    """Reads fileobject as DST file"""
    if pattern is None:
        pattern = EmbPattern()
    return read_embroidery(DstReader, f, pattern)


def read_pec(f, pattern=None):
    """Reads fileobject as PEC file"""
    if pattern is None:
        pattern = EmbPattern()
    return read_embroidery(PecReader, f, pattern)


def read_pes(f, pattern=None):
    """Reads fileobject as PES file"""
    if pattern is None:
        pattern = EmbPattern()
    return read_embroidery(PesReader, f, pattern)


def read_exp(f, pattern=None):
    """Reads fileobject as EXP file"""
    if pattern is None:
        pattern = EmbPattern()
    return read_embroidery(ExpReader, f, pattern)


def read_vp3(f, pattern=None):
    """Reads fileobject as VP3 file"""
    if pattern is None:
        pattern = EmbPattern()
    return read_embroidery(Vp3Reader, f, pattern)


def read_jef(f, pattern=None):
    """Reads fileobject as JEF file"""
    if pattern is None:
        pattern = EmbPattern()
    return read_embroidery(JefReader, f, pattern)


def read(filename, pattern=None):
    """Reads file, assuming type by extention"""
    if pattern is None:
        pattern = EmbPattern()
    extension = get_extension_by_filename(filename)
    extension = extension.lower()
    try:
        with open(filename, "rb") as f:
            if extension == "dst":
                return read_dst(f, pattern)
            elif extension == "pec":
                return read_pec(f, pattern)
            elif extension == "pes":
                return read_pes(f, pattern)
            elif extension == "exp":
                return read_exp(f, pattern)
            elif extension == "vp3":
                return read_vp3(f, pattern)
            elif extension == "jef":
                return read_jef(f, pattern)
    except IOError:
        pass


def write_embroidery(writer, pattern, stream, encode_settings=None):
    if encode_settings is None:
        encode_settings = {}
    else:
        encode_settings = encode_settings.copy()
    if not ("max_jump" in encode_settings):
        encode_settings["max_jump"] = writer.MAX_JUMP_DISTANCE
    if not ("max_stitch" in encode_settings):
        encode_settings["max_stitch"] = writer.MAX_STITCH_DISTANCE
    if isinstance(stream, str):
        with open(stream, "wb") as stream:
            normalpattern = pattern.get_normalized_pattern(encode_settings)
            writer.write(normalpattern, stream)
    else:
        normalpattern = pattern.get_normalized_pattern(encode_settings)
        writer.write(normalpattern, stream, encode_settings)


def write_dst(pattern, stream, encode_settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(DstWriter, pattern, stream, encode_settings)


def write_pec(pattern, stream, encode_settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(PecWriter, pattern, stream, encode_settings)


def write_pes(pattern, stream, encode_settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(PesWriter, pattern, stream, encode_settings)


def write_exp(pattern, stream, encode_settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(ExpWriter, pattern, stream, encode_settings)


def write_vp3(pattern, stream, encode_settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(Vp3Writer, pattern, stream, encode_settings)


def write_jef(pattern, stream, encode_settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(JefWriter, pattern, stream, encode_settings)


def write_svg(pattern, stream, encode_settings=None):
    """Writes fileobject as DST file"""
    write_embroidery(SvgWriter, pattern, stream, encode_settings)


def write(pattern, filename, encode_settings=None):
    """Writes file, assuming type by extention"""
    extension = get_extension_by_filename(filename)
    extension = extension.lower()
    with open(filename, "wb") as stream:
        if extension == "dst":
            write_dst(pattern, stream, encode_settings)
        elif extension == "pec":
            write_pec(pattern, stream, encode_settings)
        elif extension == "pes":
            write_pes(pattern, stream, encode_settings)
        elif extension == "exp":
            write_exp(pattern, stream, encode_settings)
        elif extension == "vp3":
            write_vp3(pattern, stream, encode_settings)
        elif extension == "jef":
            write_jef(pattern, stream, encode_settings)
        elif extension == "svg":
            write_svg(pattern, stream, encode_settings)
        else:
            pass
        stream.close()



import math

from .EmbConstant import *
from .WriteHelper import write_string_utf8

MAX_JUMP_DISTANCE = 121
MAX_STITCH_DISTANCE = 121
PPMM = 10
DSTHEADERSIZE = 512


def bit(b):
    return 1 << b


def encode_record(x, y, flags):
    y = -y  # flips the coordinate y space.
    b0 = 0
    b1 = 0
    b2 = 0
    if flags is JUMP:
        b2 += bit(7)  # jumpstitch 10xxxx11
    if flags is STITCH or flags is JUMP:
        b2 += bit(0)
        b2 += bit(1)
        if x > 40:
            b2 += bit(2)
            x -= 81
        if x < -40:
            b2 += bit(3)
            x += 81
        if x > 13:
            b1 += bit(2)
            x -= 27
        if x < -13:
            b1 += bit(3)
            x += 27
        if x > 4:
            b0 += bit(2)
            x -= 9
        if x < -4:
            b0 += bit(3)
            x += 9
        if x > 1:
            b1 += bit(0)
            x -= 3
        if x < -1:
            b1 += bit(1)
            x += 3
        if x > 0:
            b0 += bit(0)
            x -= 1
        if x < 0:
            b0 += bit(1)
            x += 1
        if x != 0:
            pass  # this might actually want to raise an error, if true, x was > 121 to start
        if y > 40:
            b2 += bit(5)
            y -= 81
        if y < -40:
            b2 += bit(4)
            y += 81
        if y > 13:
            b1 += bit(5)
            y -= 27
        if y < -13:
            b1 += bit(4)
            y += 27
        if y > 4:
            b0 += bit(5)
            y -= 9
        if y < -4:
            b0 += bit(4)
            y += 9
        if y > 1:
            b1 += bit(7)
            y -= 3
        if y < -1:
            b1 += bit(6)
            y += 3
        if y > 0:
            b0 += bit(7)
            y -= 1
        if y < 0:
            b0 += bit(6)
            y += 1
        if y != 0:
            pass  # Fail. y > 121
    elif flags is COLOR_CHANGE:
        b2 = 0b11000011
    elif flags is STOP:
        b2 = 0b11000011
    elif flags is END:
        b2 = 0b11110011
    elif flags is SEQUIN:
        b2 = 0b01000011
    return bytes(bytearray([b0, b1, b2]))


def write(pattern, f, settings=None):
    extended_header = False
    if settings is not None:
        extended_header = settings.get("extended header", extended_header)

    extends = pattern.extends()
    width = extends[2] - extends[0]
    height = extends[3] - extends[1]

    name = pattern.get_metadata("name", "Untitled")

    write_string_utf8(f,"LA:%-16s\r" % name)
    write_string_utf8(f,"ST:%7d\r" % pattern.count_stitches())
    write_string_utf8(f,"CO:%3d\r" % pattern.count_color_changes())
    x_extend = math.ceil(PPMM * width / 2)
    y_extend = math.ceil(PPMM * height / 2)
    write_string_utf8(f,"+X:%5d\r" % x_extend)
    write_string_utf8(f,"-X:%5d\r" % x_extend)
    write_string_utf8(f,"+Y:%5d\r" % y_extend)
    write_string_utf8(f,"-Y:%5d\r" % y_extend)
    write_string_utf8(f,"AX:+%5d\r" % 0)
    write_string_utf8(f,"AY:+%5d\r" % 0)
    write_string_utf8(f,"MX:+%5d\r" % 0)
    write_string_utf8(f,"AY:+%5d\r" % 0)
    write_string_utf8(f,"PD:%6s\r" % "******")
    if extended_header:
        author = pattern.get_metadata("author")
        if author is not None:
            write_string_utf8(f,"AU:%s\r" % author)
        meta_copyright = pattern.get_metadata("copyright")
        if meta_copyright is not None:
            write_string_utf8(f,"CP:%s\r" % meta_copyright)
        if len(pattern.threadlist) > 0:
            for thread in pattern.threadlist:
                write_string_utf8(f,"TC:%s,%s,%s\r" %
                                  (thread.hex_color(),
                                   thread.description,
                                   thread.catalog_number))
    f.write(b'\x1a')
    for i in range(f.tell(), DSTHEADERSIZE):
        f.write(b'\x20')  # space

    stitches = pattern.stitches
    xx = 0
    yy = 0
    for stitch in stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2]
        dx = x - xx
        dy = y - yy
        if data == TRIM:
            f.write(encode_record(2, 2, JUMP))
            f.write(encode_record(-4, -4, JUMP))
            f.write(encode_record(2, 2, JUMP))
        else:
            f.write(encode_record(round(dx), round(dy), data))
        xx = x
        yy = y



from .EmbConstant import *

MAX_JUMP_DISTANCE = 127
MAX_STITCH_DISTANCE = 127


def write(pattern, f, settings=None):
    stitches = pattern.stitches
    jumping = False
    xx = 0
    yy = 0
    for stitch in stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2]
        dx = x - xx
        dy = y - yy
        if data is STITCH:
            if jumping:
                f.write(b'\x00\x00')
                jumping = False
            delta_x = int(round(dx)) & 0xFF
            delta_y = -int(round(dy)) & 0xFF
            f.write(bytes(bytearray([delta_x, delta_y])))
        elif data == JUMP:
            jumping = True
            delta_x = int(round(dx)) & 0xFF
            delta_y = -int(round(dy)) & 0xFF
            f.write(b'\x80\x04')
            f.write(bytes(bytearray([delta_x, delta_y])))
        elif data == TRIM:
            f.write(b'\x80\x80\x07\x00')
            f.write(b'\x80\x04\x00\x00')
            continue
        elif data == COLOR_CHANGE:
            if jumping:
                f.write(b'\x00\x00')
                jumping = False
            f.write(b'\x80\x01\x00\x00')
            f.write(b'\x80\x80\x07\x00')
            continue
        elif data == STOP:
            if jumping:
                f.write(b'\x00\x00')
                jumping = False
            f.write(b'\x80\x01\x00\x00')
            continue
        elif data == END:
            pass
        xx = x
        yy = y



from .EmbThreadJef import get_thread_set
from .ReadHelper import read_signed, read_int_32le, signed


def read(f, read_object):
    jef_threads = get_thread_set()
    stitch_offset = read_int_32le(f)
    f.seek(20, 1)
    count_colors = read_int_32le(f)
    count_stitches = read_int_32le(f)
    f.seek(84, 1)
    for i in range(0, count_colors):
        index = abs(read_int_32le(f))
        read_object.add_thread(jef_threads[index % 79])

    f.seek(stitch_offset - 116 - (count_colors * 4), 1)
    for i in range(0, count_stitches + 100):
        b = read_signed(f, 2)
        if (b[0] & 0xFF) == 0x80:
            if (b[1] & 1) != 0:
                b = read_signed(f, 2)
                read_object.color_change(0, 0)
                read_object.move(b[0], -b[1])
            elif b[1] == 0x04 or b[1] == 0x02:  # trim
                b = read_signed(f, 2)
                read_object.trim(0, 0)
                read_object.move(b[0], -b[1])
            elif b[1] == 0x10:  # end
                break
        else:
            read_object.stitch(signed(b[0]),
                               -signed(b[1]))
    read_object.end(0, 0)



import math

from .EmbConstant import *


class Transcoder:
    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        self.translate_x = settings.get("translate_x", 0)
        self.translate_y = settings.get("translate_y", 0)
        self.has_tie_on = settings.get("tie_on", False)
        self.has_tie_off = settings.get("tie_off", False)
        self.max_stitch = settings.get("max_stitch", float('inf'))
        self.max_jump = settings.get("max_jump", float('inf'))
        self.jump_threshold = settings.get("jump_threshold", self.max_jump)
        self.source_pattern = None
        self.dest_pattern = None
        self.position = 0
        self.color_index = -1
        self.stitch = None
        self.state_trimmed = True
        self.needle_x = 0
        self.needle_y = 0
        self.subdivide_long_stitches = self.max_stitch < self.jump_threshold

    def transcode(self, source_pattern, destination_pattern):
        self.source_pattern = source_pattern
        self.dest_pattern = destination_pattern
        self.transcode_metadata()
        self.transcode_threads()
        self.transcode_stitches()
        return destination_pattern

    def transcode_metadata(self):
        """Transcodes metadata, (just moves)"""
        source = self.source_pattern
        dest = self.dest_pattern
        self.dest_pattern.extras = self.source_pattern.extras.copy()

    def transcode_threads(self):
        """Transcodes threads, (just moves)"""
        source = self.source_pattern.threadlist
        dest = self.dest_pattern.threadlist
        dest.extend(source)

    def transcode_stitches(self):
        """Transcodes stitches.
        Converts middle-level commands and potentially incompatible
        commands into a format friendly low level commands."""
        source = self.source_pattern.stitches
        self.state_trimmed = True
        self.needle_x = 0
        self.needle_y = 0
        self.position = 0
        self.color_index = -1
        flags = NO_COMMAND

        for self.position, self.stitch in enumerate(source):
            x = round(self.stitch[0] + self.translate_x)
            y = round(self.stitch[1] + self.translate_y)
            flags = self.stitch[2]

            if flags == NO_COMMAND:
                continue

            elif flags == STITCH:
                if self.state_trimmed:
                    self.jump_to(x, y)
                    if self.has_tie_on:
                        self.tie_on()
                    self.stitch_at(x, y)
                else:
                    if self.subdivide_long_stitches:
                        self.sew_to(x, y)
                    else:
                        self.needle_to(x, y)
            elif flags == NEEDLE_AT:
                if self.state_trimmed:
                    self.jump_to(x, y)
                    if self.has_tie_on:
                        self.tie_on()
                    self.stitch_at(x, y)
                else:
                    self.needle_to(x, y)
            elif flags == SEW_TO:
                if self.state_trimmed:
                    self.jump_to(x, y)
                    if self.has_tie_on:
                        self.tie_on()
                    self.stitch_at(x, y)
                else:
                    self.sew_to(x, y)

            # Middle Level Commands.
            elif flags == FRAME_EJECT:
                self.tie_off_and_trim_if_needed()
                self.jump_to(x, y)
                self.stop_here()
            elif flags == SEQUENCE_BREAK:
                self.tie_off_and_trim_if_needed()
            elif flags == COLOR_BREAK:
                self.tie_off_and_trim_if_needed()
                self.color_change_here_if_needed()
            elif flags == TIE_OFF:
                self.tie_off()
            elif flags == TIE_ON:
                self.tie_on()

            # Core Commands.
            elif flags == TRIM:
                self.tie_off_and_trim_if_needed()
            elif flags == JUMP:
                self.jump_to(x, y)
            elif flags == SEQUIN:
                # While DST are the only files with this there are some
                # chances that we might need to further alter the routines
                # built here by the format. For example dst calls a sequin
                # command out of jumps and toggles that on and off. Other
                # formats might use a different paradigm.
                if self.state_trimmed:
                    self.jump_to(x, y)
                    if self.has_tie_on:
                        self.tie_on()
                else:
                    self.sew_to(x, y)
                self.sequin_at(x, y)
            elif flags == COLOR_CHANGE:
                self.tie_off_and_trim_if_needed()
                self.color_change_here()
                # If we are told to do something we do it.
                # Even if it's the first command and makes no sense.
            elif flags == STOP:
                self.stop_here()
            elif flags == END:
                self.end_here()
                break

            # On-the-fly Settings Commands.
            elif flags == ENABLE_TIE_ON:
                self.has_tie_on = True
            elif flags == ENABLE_TIE_OFF:
                self.has_tie_off = True
            elif flags == DISABLE_TIE_ON:
                self.has_tie_on = False
            elif flags == DISABLE_TIE_OFF:
                self.has_tie_off = False
            elif flags == TRANSLATE:
                x = self.stitch[0]
                y = self.stitch[1]
                self.translate_x += x
                self.translate_y += y
            elif flags == JUMP_THRESHOLD:
                x = self.stitch[0]
                self.jump_threshold = x
                self.subdivide_long_stitches = self.max_stitch < self.jump_threshold
            elif flags == MAX_JUMP_LENGTH:
                x = self.stitch[0]
                self.max_jump = x
            elif flags == MAX_STITCH_LENGTH:
                x = self.stitch[0]
                self.max_stitch = x
                self.subdivide_long_stitches = self.max_stitch < self.jump_threshold
        if flags != END:
            self.end_here()

    def update_needle_position(self, x, y):
        self.needle_x = x
        self.needle_y = y

    def declare_not_trimmed(self):
        if self.state_trimmed:
            self.state_trimmed = False
            if self.color_index == -1:
                self.color_index = 0

    def add(self, flags, x=None, y=None):
        if x is None:
            x = self.needle_x
        if y is None:
            y = self.needle_y
        self.dest_pattern.stitches.append([x, y, flags])

    def tie_off_and_trim_if_needed(self):
        if not self.state_trimmed:
            self.tie_off_and_trim()

    def tie_off_and_trim(self):
        if self.has_tie_off:
            self.tie_off()
        self.trim_here()

    def tie_off(self):
        try:
            b = self.source_pattern.stitches[self.position - 1]
            x = round(b[0] + self.translate_x)
            y = round(b[1] + self.translate_y)
            flags = self.stitch[2]
            if flags == STITCH or flags == NEEDLE_AT or flags == SEW_TO:
                self.lock_stitch(self.needle_x, self.needle_y,
                                 x, y, self.max_stitch)
        except IndexError:
            pass  # must be an island stitch. jump-stitch-jump

    def tie_on(self):
        try:
            b = self.source_pattern.stitches[self.position + 1]
            x = round(b[0] + self.translate_x)
            y = round(b[1] + self.translate_y)
            flags = self.stitch[2]
            if flags == STITCH or flags == NEEDLE_AT or flags == SEW_TO:
                self.lock_stitch(self.needle_x, self.needle_y,
                                 x, y, self.max_stitch)
        except IndexError:
            pass  # must be an island stitch. jump-stitch-jump

    def trim_here(self):
        self.add(TRIM)
        self.state_trimmed = True

    def jump_to(self, new_x, new_y):
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_jump
        self.constrained_step_to(x0, y0, new_x, new_y, max_length, JUMP)
        self.add(JUMP, new_x, new_y)
        self.update_needle_position(new_x, new_y)

    def sew_to(self, new_x, new_y):
        """Stitches to a specific location, with the emphasis on sewing.
         Subdivides long stitches into additional stitches.
        """
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_stitch
        self.constrained_step_to(x0, y0, new_x, new_y, max_length, STITCH)
        self.stitch_at(new_x, new_y)

    def needle_to(self, new_x, new_y):
        """Insert needle at specific location, emphasis on the needle.
        Uses jumps to avoid needle penetrations where possible.

        The limit here is the max stitch limit or jump threshold.
        If jump threshold is set low, it will insert jumps even
        between stitches it could have technically encoded values for.

        Stitches to the new location, adding jumps if needed.
        """
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = min(self.jump_threshold, self.max_stitch)
        if self.position_will_exceed_constraint(max_length):
            pass
        self.constrained_step_to(x0, y0, new_x, new_y, max_length, JUMP)
        self.stitch_at(new_x, new_y)

    def stitch_at(self, new_x, new_y):
        """Inserts a stitch at the specific location.
        Should have already been checked for constraints."""
        self.add(STITCH, new_x, new_y)
        self.update_needle_position(new_x, new_y)
        self.declare_not_trimmed()

    def sequin_at(self, new_x, new_y):
        # TODO: There might be other middle-level commands needed here.
        self.add(SEQUIN)
        self.update_needle_position(new_x, new_y)
        self.declare_not_trimmed()

    def stop_here(self):
        self.add(STOP)
        self.state_trimmed = True

    def end_here(self):
        self.add(END)
        self.state_trimmed = True

    def color_change_here_if_needed(self):
        if self.color_index >= 0:  # Have we stitched anything yet?
            self.color_change_here()
            # We should actually look ahead and ensure
            # there are no more objects that will become stitches.
            # post-stitch color-changes are pointless.

    def color_change_here(self):
        self.add(COLOR_CHANGE)
        self.color_index += 1
        self.state_trimmed = True

    def position_will_exceed_constraint(self, length=None, new_x=None, new_y=None):
        """Check if the stitch is too long before trying to deal with it."""
        if length is None:
            length = self.max_stitch
        if new_x is None or new_y is None:
            new_x = round(self.stitch[0] + self.translate_x)
            new_y = round(self.stitch[1] + self.translate_y)
        distance_x = new_x - self.needle_x
        distance_y = new_y - self.needle_y
        return abs(distance_x) > length or abs(distance_y) > length

    def constrained_step_to(self, x0, y0, x1, y1, max_length, data):
        """Command sequence line to x, y, respecting length as maximum.
        This does not arrive_at, it steps to within striking distance.
        The next step can arrive at (x, y) without violating constraint.
        If these are already in range, this command will do nothing.
        """
        transcode = self.dest_pattern.stitches
        distance_x = x1 - x0
        distance_y = y1 - y0
        if abs(distance_x) > max_length or abs(distance_y) > max_length:
            steps_x = math.ceil(abs(distance_x / max_length))
            steps_y = math.ceil(abs(distance_y / max_length))
            if steps_x > steps_y:
                steps = steps_x
            else:
                steps = steps_y
            step_size_x = distance_x / steps
            step_size_y = distance_y / steps

            q = 0
            qe = steps
            qx = x0
            qy = y0
            while q < qe:
                transcode.append([round(qx), round(qy), data])
                q += 1
                qx += step_size_x
                qy += step_size_y

    def lock_stitch(self, x, y, anchor_x, anchor_y, max_length=None):
        """Tie-on, Tie-off. Lock stitch from current location towards
        anchor location.Ends again at lock location. May not exceed
        max_length in the process."""
        if max_length is None:
            max_length = self.max_stitch
        transcode = self.dest_pattern.stitches
        dist = distance(x, y, anchor_x, anchor_y)
        if dist > max_length:
            p = oriented(x, y, anchor_x, anchor_y, max_length)
            anchor_x = p[0]
            anchor_y = p[1]
        for amount in (0, .33, .66, .33, 0):
            transcode.append([
                towards(x, anchor_x, amount),
                towards(y, anchor_y, amount),
                STITCH])


def distance_squared(x0, y0, x1, y1):
    """squared of distance between x0,y0 and x1,y1"""
    dx = x1 - x0
    dy = y1 - y0
    dx *= dx
    dy *= dy
    return dx + dy


def distance(x0, y0, x1, y1):
    """distance between x0,y0 and x1,y1"""
    return math.sqrt(distance_squared(x0, y0, x1, y1))


def towards(a, b, amount):
    """amount between [0,1] -> [a,b]"""
    return (amount * (b - a)) + a


def angle_radians(x0, y0, x1, y1):
    """Angle in radians between x0,y0 and x1,y1"""
    return math.atan2(y1 - y0, x1 - x0)


def oriented(x0, y0, x1, y1, r):
    """from x0,y0 in the direction of x1,y1 in the distance of r"""
    radians = angle_radians(x0, y0, x1, y1)
    return x0 + (r * math.cos(radians)), y0 + (r * math.sin(radians))




def signed(b):
    if b > 127:
        return -256 + b
    else:
        return b


def read(f, read_object):
    while True:
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if b[0] & 0xFF == 0x80:
            if b[1] == 0x80:  # Trim
                b = bytearray(f.read(2))
                if len(b) != 2:
                    break
                read_object.trim(0, 0)
            elif b[1] == 0x02:
                read_object.stitch(signed(b[0]), -(signed(b[1])))
            elif b[1] == 0x04:  # Jump
                b = bytearray(f.read(2))
                if len(b) != 2:
                    break
                read_object.move(signed(b[0]), -signed(b[1]))
            elif b[1] == 0x01:  # Colorchange
                b = bytearray(f.read(2))
                if len(b) != 2:
                    break
                read_object.color_change(0, 0)
                read_object.move(signed(b[0]), -signed(b[1]))
        else:
            read_object.stitch(signed(b[0]), -signed(b[1]))



NO_COMMAND = -1
STITCH = 0
JUMP = 1
TRIM = 2
STOP = 3
END = 4
COLOR_CHANGE = 5
SEQUIN = 6

# Different stitch classes.
# If the stitch is long? How do we get there?
SEW_TO = 0xB0
NEEDLE_AT = 0xB1

SEQUENCE_BREAK = 0xE1
COLOR_BREAK = 0xE2
TIE_ON = 0xE4
TIE_OFF = 0xE5
FRAME_EJECT = 0xE9

TRANSLATE = 0xD0
ENABLE_TIE_ON = 0xD1
ENABLE_TIE_OFF = 0xD2
DISABLE_TIE_ON = 0xD3
DISABLE_TIE_OFF = 0xD4
MAX_STITCH_LENGTH = 0xD5
MAX_JUMP_LENGTH = 0xD6
JUMP_THRESHOLD = 0xD8


# Eventually the commands are supposed to be limited to 255 thereby
# allowing additional information like, color change to color in position n
# To be stored in the higher level bits.
COMMAND_MASK = 0xFF
